/*
	GIER console (Motif)

	(C) Copyright 2011 by Mogens Kjaer


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <math.h>
#include <sys/ioctl.h>
#include <termios.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>
#include <Xm/Protocols.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include <X11/xpm.h>
#include <X11/keysym.h>

/* For motif 1.2 */

#ifndef XmUNSET
#define XmUNSET False
#endif
#ifndef XmSET
#define XmSET True
#endif

#define ABS(x) ( ((x)<0) ? (-(x)) : (x) )

#include "console.h"

#include "typewriter.h"

static XtAppContext app;


static Widget toplevel;
static Widget typewriter_wnd=NULL, reader_wnd=NULL, punch_wnd=NULL, printer_wnd=NULL;

int n;
Arg wargs[1000];

/*
	External functions:
*/

extern unsigned char flx2a(char, int *);
extern unsigned char a2flx(char, int *);
extern unsigned char remove_parity(unsigned char);
extern int ps_open(char *, int, double, int, int, int);
extern int ps_SY(unsigned char);
extern void ps_close();

/*
	Local variables:
*/

/*Main*/
Widget main_wnd;
Widget main_viewmenu, main_inputmenu, main_outputmenu;
static int OKPressed, CancelPressed, Wnd_destroyed, debug_active=0;
/*typewriter*/
static Widget typewriter_form, typewriter_swindow, typewriter_horScroll,
       typewriter_verScroll, typewriter_input_label, typewriter_drawing;
static int typewriter_drawing_width, typewriter_drawing_height;
static short paper[LINEBUFFER][PAPERWIDTH][MAXOVERWRITE];
static int cur_row, cur_col;
static int uppercase; /* 0: lower, 1: upper */
static int old_uppercase;
static int redtext; /* 0: black, 1: red */
static int typewriter_font_width, typewriter_font_height;
static int in_copy=0;
static GC black_normal_gc;
static GC red_normal_gc;
static GC black_symbol_gc;
static GC red_symbol_gc;
static GC black_subscript_gc;
static GC red_subscript_gc;
static GC gc_table[2][3]; /* [redtext][0: normal, 1: symbol, 2: subscript] */
static GC scroll_gc;
static int typewriter_waiting=0;
static XComposeStatus compose;
static char strbuf[100];
#define RED_MASK 256
#define SYMBOL_MASK 512
#define typewriter_CursorX(col) ( (col)*typewriter_font_width )
#define typewriter_CursorY(row) ( (row)*typewriter_font_height + typewriter_applicationdata.textfont->ascent )
/*reader*/
static Widget reader_form, reader_reset, reader_pread, reader_skip, reader_up,
              reader_drawing, reader_filedialog;
static int reader_drawing_width, reader_drawing_height;
static int reader_current_fd = -1;
static int reader_modified = False;
static int reader_filetype; /* 0: flx, 1: asc */
static int reader_uppercase=0;
static int reader_nsavedchars=0;
static int reader_linepos=0;
static unsigned char reader_savedchars[10];
static int reader_font_width, reader_font_height;
#define reader_CursorX(col) ( (col)*reader_font_width )
#define reader_CursorY(row) ( (row)*reader_font_height + reader_applicationdata.textfont->ascent)
static GC reader_hole_gc, reader_gray_gc;
/*printer*/
static Widget printer_form, printer_newfile, printer_endfile, printer_psfilename;
static Widget printer_filedialog, printer_fileform, printer_fileselection;
static Widget papersize, lpi_edit, fontsize_edit, lines_edit, uadvance_select, uadvance_check;
static char lpi_buffer[20];
static char fontsize_buffer[20];
static char lines_buffer[20];
static int is_printing=0;
static int letter=0, lpi=6, lines=67;
Boolean uadvance=True;
static double fontsize=11.0;

/*punch*/
static Widget punch_form, punch_newtape, punch_endtape, punch_tapename, punch_filedialog=NULL, punch_drawing;
static int punch_drawing_width, punch_drawing_height;
static int punch_current_fd = -1;
static int punch_filetype; /* 0: flx, 1: asc */
static int punch_uppercase=0;
static int punch_font_width, punch_font_height;
#define punch_CursorX(col) ( (col)*punch_font_width )
#define punch_CursorY(row) ( (row)*punch_font_height + punch_applicationdata.textfont->ascent)
static GC punch_hole_gc, punch_gray_gc;
static char punch_char_table[PUNCH_TAPE_WINDOW];

/*
	Forward declarations:
*/
void GIER_run();
void GIER_start();
void interface_update_all();
static void typewriter_copywait();
void typewriter_SY(char);
void typewriter_wait(int);
void typewriter_lowercase();
void typewriter_init();
void typewriter_destroy();
int typewriter_is_visible();
void reader_update();
void reader_wait(int status);
void reader_init();
void reader_destroy();
int reader_is_visible();
void printer_init();
static void printer_fileopen_callback(Widget, XtPointer, XtPointer);
static void printer_fileclose_callback(Widget, XtPointer, XtPointer);
void printer_SY(char);
void printer_destroy();
int printer_is_visible();
void punch_SY(char);
void punch_init();
void punch_destroy();
int punch_is_visible();

void GIER_exit()
{
  if(is_printing) ps_close();
  exit(0);
}

static void maindestroy_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  GIER_exit();
}

static void quit_action()
{
  GIER_exit();
}

static int old_status=0;

void YE_status()
{
  XmString stmp;

  if( (old_status != YE_wait) && typewriter_is_visible())
  {
    n=0;
    if(YE_wait)
    {
      XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("INPUT"))); n++;
    }
    else
    {
      XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("     "))); n++;
    }
    XtSetValues(typewriter_input_label, wargs, n);
    XmStringFree(stmp);
    old_status=YE_wait;
  }
}

static void filemenu_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Cardinal item_no = (long) call_data;
  switch(item_no)
  {
  case 0: /* Quit */
    quit_action();
    break;
  }
}

void update_viewmenu()
{
  XtVaSetValues(XtNameToWidget(main_viewmenu, "button_0"), XmNset, typewriter_wnd!=NULL, NULL);
  XtVaSetValues(XtNameToWidget(main_viewmenu, "button_1"), XmNset, reader_wnd!=NULL, NULL);
  XtVaSetValues(XtNameToWidget(main_viewmenu, "button_2"), XmNset, punch_wnd!=NULL, NULL);
  XtVaSetValues(XtNameToWidget(main_viewmenu, "button_3"), XmNset, printer_wnd!=NULL, NULL);
}

static void viewmenu_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Cardinal item_no = (long) call_data;
  XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) client_data;

  switch(item_no)
  {
  case 0: /* Typewriter */
    if(cbs->set)
    {
      typewriter_init();
    }
    else
    {
      typewriter_destroy();
    }
    break;
  case 1: /* Tape reader */
    if(cbs->set)
    {
      reader_init();
    }
    else
    {
      reader_destroy();
    }
    break;
  case 2: /* Tape punch */
    if(cbs->set)
    {
      punch_init();
    }
    else
    {
      punch_destroy();
    }
    break;
  case 3: /* Line printer */
    if(cbs->set)
    {
      printer_init();
    }
    else
    {
      printer_destroy();
    }
    break;
  }
}

static void inputmenu_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Cardinal item_no = (long) call_data;
  XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) client_data;
  switch(item_no)
  {
    case 0: /* typewriter */
      BY &= ~1;
      if(cbs->set) BY |= 1;
      break;
    case 1: /* reader */
      BY &= ~2;
      if(cbs->set) BY |= 2;
      break;
  }
  if(YE_wait)
  {
    if(usb_send_input("From inputmenu_callback"))
    {
      YE_wait=0;
      YE_status();
    }
  }
}

static void outputmenu_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Cardinal item_no = (long) call_data;
  XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) client_data;
  switch(item_no)
  {
    case 0: /* line printer */
      BY &= ~8;
      if(cbs->set) BY |= 8;
      break;
    case 1: /* typewriter */
      BY &= ~16;
      if(cbs->set) BY |= 16;
      break;
    case 2: /* punch */
      BY &= ~32;
      if(cbs->set) BY |= 32;
      break;
  }
}

void main_wnd_init()
{
  Atom atom;
  Widget menubar, filemenu;
  /*
  	This is the main window of the GIER console simulator.

  */
  Widget new_form;
  XmString s[20];
  char tmp_label[80];
  char tmp_title[80];
  int i, j;

  n=0;
  XtSetArg(wargs[n], XtNtitle, "GIER Console"); n++;
  XtSetArg(wargs[n], XtNgeometry, "+10+10"); n++;
  main_wnd = XmCreateMainWindow(toplevel, "main", wargs, n);
  atom = XmInternAtom(XtDisplay(toplevel),"WM_DELETE_WINDOW",TRUE);
  XmAddWMProtocolCallback(toplevel, atom, maindestroy_callback, False);
  XtManageChild(main_wnd);

  /* Menu bar */

  menubar = XmVaCreateSimpleMenuBar(main_wnd, "menubar",
              XmVaCASCADEBUTTON, s[0] = XmStringCreateSimple("File"), (KeySym) 'F',     /* File         */
	      XmVaCASCADEBUTTON, s[1] = XmStringCreateSimple("View"), (KeySym) 'V',	/* View		*/
	      XmVaCASCADEBUTTON, s[2] = XmStringCreateSimple("Input"), (KeySym) 'I',	/* Input	*/
	      XmVaCASCADEBUTTON, s[3] = XmStringCreateSimple("Ouput"), (KeySym) 'O',	/* Output	*/
	      NULL);
  for(i=0; i<4; i++) XmStringFree(s[i]);

  filemenu = XmVaCreateSimplePulldownMenu(menubar, "filemenu", 0, filemenu_callback,
              XmVaPUSHBUTTON, s[0] = XmStringCreateSimple("Quit"), (KeySym) 'Q', NULL, NULL,    /* Quit         */
	      NULL);
  for(i=0; i<1; i++) XmStringFree(s[i]);

  main_viewmenu = XmVaCreateSimplePulldownMenu(menubar, "viewmenu", 1, viewmenu_callback,
              XmVaCHECKBUTTON, s[0] = XmStringCreateSimple("View Typewriter"), (KeySym) 'T', NULL, NULL,	/* Show typewriter */
              XmVaCHECKBUTTON, s[1] = XmStringCreateSimple("View Tape Reader"), (KeySym) 'R', NULL, NULL,	/* Show tape reader */
              XmVaCHECKBUTTON, s[2] = XmStringCreateSimple("View Tape Punch"), (KeySym) 'P', NULL, NULL,	/* Show punch */
              XmVaCHECKBUTTON, s[3] = XmStringCreateSimple("View Line Printer"), (KeySym) 'L', NULL, NULL,	/* Show lineprinter */
	      NULL);
  for(i=0; i<4; i++) XmStringFree(s[i]);

  main_inputmenu = XmVaCreateSimplePulldownMenu(menubar, "inputmenu", 2, inputmenu_callback,
              XmVaCHECKBUTTON, s[0] = XmStringCreateSimple("Input Typewriter"), (KeySym) 'T', NULL, NULL,	/* Input from typewriter */
              XmVaCHECKBUTTON, s[1] = XmStringCreateSimple("Input Tape Reader"), (KeySym) 'R', NULL, NULL,	/* Input from  tape reader */
	      NULL);
  for(i=0; i<2; i++) XmStringFree(s[i]);
  XtVaSetValues(XtNameToWidget(main_inputmenu, "button_0"), XmNset, TRUE, NULL);

  main_outputmenu = XmVaCreateSimplePulldownMenu(menubar, "outputmenu", 3, outputmenu_callback,
              XmVaCHECKBUTTON, s[0] = XmStringCreateSimple("Output Printer"), (KeySym) 'L', NULL, NULL,		/* Output to line printer */
              XmVaCHECKBUTTON, s[1] = XmStringCreateSimple("Output Typewriter"), (KeySym) 'W', NULL, NULL,	/* Output to typewriter */
              XmVaCHECKBUTTON, s[2] = XmStringCreateSimple("Output Punch"), (KeySym) 'P', NULL, NULL,		/* Output to punch */
	      NULL);
  for(i=0; i<3; i++) XmStringFree(s[i]);
  XtVaSetValues(XtNameToWidget(main_outputmenu, "button_1"), XmNset, TRUE, NULL);

  XtManageChild(menubar);

  /* start everything */
  reader_init();
  printer_init();
  punch_init();
  typewriter_init();
}

void interface_init(int argc, char **argv)
{
  int local_argc;
  char **local_argv;
  int i;

  /* Make a local copy of argc/argv */

  /* Kludge... */

  local_argc = argc+4;
  local_argv=malloc(sizeof(*local_argv)*(local_argc+1));
  for(i=0;i<argc; i++) local_argv[i] = argv[i];
  local_argv[local_argc-4] = "-title";
  local_argv[local_argc-3] = "GIER Console Simulator";
  local_argv[local_argc-2] = "-geometry";
  local_argv[local_argc-1] = "+10+10";
  local_argv[local_argc] = NULL;

  toplevel=XtAppInitialize(&app,"GIER", NULL, 0, &local_argc, local_argv, NULL, NULL, 0);

  main_wnd_init();

  XtRealizeWidget(toplevel);
}

void interface_update()
{
}

#define USBDIMS "/dev/ttyACM0"
int usb_fd,usb_input_id;

void usb_send(int c)
{
  char buffer[10];
  buffer[0]=c;
  assert(write(usb_fd, &buffer, 1)==1);
  fprintf(stderr, "usb_send: %d\n", c);
}

int usb_typewriter_send()
{
  int BT,i;
  if(typewriter_nchars>0)
  {
    BT=typewriter_char_table[0];
    typewriter_nchars--;
    for(i=0; i<typewriter_nchars; i++) typewriter_char_table[i]=typewriter_char_table[i+1];
    usb_send(BT);
    return 1;
  }
  return 0;
}

int usb_reader_send()
{
  int BL,i;
  if((reader_nchars-reader_offset)>0)
  {
    BL=remove_parity((unsigned char) reader_char_table[reader_offset]);
    reader_nchars--;
    for(i=0; i<reader_nchars; i++) reader_char_table[i]=reader_char_table[i+1];
    reader_char_table[reader_nchars]=0;
    reader_wait(FALSE);
    usb_send(BL);
    return 1;
  }
  else
  {
    reader_wait(TRUE);
    return 0;
  }
}

int usb_send_input(char *txt)
{
  int sent=0;
  if(BY&1)
  {
    sent=usb_typewriter_send();
    fprintf(stderr, "usb_send_input: usb_typewriter_send: %d\n", sent);
  }
  if((!sent) && (BY&2))
  {
    sent=usb_reader_send();
    fprintf(stderr, "usb_send_input: usb_reader_send: %d\n", sent);
  }
  return sent;
}

static void usb_input_callback(XtPointer closure, int *source, XtInputId *id)
{
  int n,st;
  char buf[20],c;
  

  if (ioctl(usb_fd, FIONREAD, &n) == -1 || n == 0)
  {
    return;
  }

  n=0;
    st=read(usb_fd, &c, 1);
    if(st!=1)
    {
      return;
    }
    else
    {
      n=1;
    }
  if(n==1)
  {

    if(c == -128)
    {
      /* Turn green light on unless we already have character to send */
      if(!usb_send_input("After receiving green light"))
      {
	typewriter_wait(TRUE);
      }
    }
    else if(c == -127)
    {
    }
    else
    {
      fprintf(stderr, "print char: %d\n", c);
      if(BY&8) printer_SY(c);
      if(BY&16) typewriter_SY(c);
      if(BY&32) punch_SY(c);
    }
  }
  else
  {
  }
}

static void setup_input()
{
  struct termios t;
  char c;

  assert((usb_fd = open(USBDIMS, O_RDWR))>=0);

  assert(tcgetattr(usb_fd, &t)==0);
  cfmakeraw(&t);
  cfsetspeed(&t, B38400);
  cfsetspeed(&t, B115200);
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 2;
  t.c_cflag |= CLOCAL;
  t.c_cflag &= ~CRTSCTS;
  t.c_iflag &= ~IXOFF;
  t.c_iflag &= ~IXON;
  assert(tcsetattr(usb_fd, TCSAFLUSH, &t) == 0);
  t.c_cc[VTIME] = 10;
  assert(tcsetattr(usb_fd, TCSAFLUSH, &t) == 0);
  
  usb_input_id = XtAppAddInput(app, usb_fd, (XtPointer) XtInputReadMask, usb_input_callback, NULL);
}

void interface_run()
{
  XEvent event;
  int status;

  setup_input();

AGAIN:

  status=XtAppPending(app);

  if(status)
  {
AGAIN2:
    XtAppProcessEvent(app, status);
    goto AGAIN;
  }
  reader_update();
  if(YE_wait)
  {
    if(usb_send_input("From main event loop"))
    {
      YE_wait=0;
      YE_status();
    }
  }
  if(!YE_wait) goto AGAIN;
  status=XtIMAll;
  goto AGAIN2;
}

void interface_update_all()
{
  update_viewmenu();
}

/* Local variables */

typedef struct
{
  XFontStruct *textfont;
  XFontStruct *subscriptfont;
  XFontStruct *symbolfont;
  Pixel lampcolor, redcolor, blackcolor;

} typewriter_ApplicationData, *typewriter_ApplicationDataPtr;

static typewriter_ApplicationData typewriter_applicationdata;

static XtResource typewriter_resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(typewriter_ApplicationDataPtr, textfont),
   XtRString, "-adobe-courier-medium-r-*-*-*-180-75-75-*-*-*-*"
  },
  {"subscriptfont", "Subscriptfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(typewriter_ApplicationDataPtr, subscriptfont),
   XtRString, "-adobe-courier-medium-r-*-*-*-100-75-75-*-*-*-*"
  },
  {"symbolfont", "Symbolfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(typewriter_ApplicationDataPtr, symbolfont),
   XtRString, "-adobe-symbol-medium-r-*-*-*-180-75-75-*-*-*-*"
  },
  {"lampcolor", "Lampcolor", XtRPixel, sizeof(Pixel),
   XtOffset(typewriter_ApplicationDataPtr, lampcolor),
   XtRString, "green"
  },
  {"redcolor", "Redcolor", XtRPixel, sizeof(Pixel),
   XtOffset(typewriter_ApplicationDataPtr, redcolor),
   XtRString, "red"
  },
  {"blackcolor", "Blackcolor", XtRPixel, sizeof(Pixel),
   XtOffset(typewriter_ApplicationDataPtr, blackcolor),
   XtRString, "black"
  },
}; 

static void typewriter_drawchar(int row, int col)
{
  short c;
  int overprint;
  Position x,y;
  Dimension width, height;
  char text[3];
  int tlength;
  int ired, ifont;

  overprint=0;

  x = typewriter_CursorX(col);
  y = typewriter_CursorY(row);

  XDrawImageString(XtDisplay(typewriter_drawing), XtWindow(typewriter_drawing), gc_table[0][0], x, y, " ", 1);

  while( (overprint<MAXOVERWRITE) &&
         (c=paper[row][col][overprint++]) != 0)
  {
    ired = (c&RED_MASK)!=0;
    ifont = (c&SYMBOL_MASK)!=0;
    text[0] = (char) (c&255);
    tlength=1;

    if(text[0]=='\'')
    {
      /* subscript 10 */
      ifont = 2;
      text[0] = '1';
      text[1] = '0';
      tlength=2;

      y=y+(typewriter_applicationdata.textfont->descent-typewriter_applicationdata.subscriptfont->descent);
    }

    if(overprint == 1)
      XDrawImageString(XtDisplay(typewriter_drawing), XtWindow(typewriter_drawing), gc_table[ired][ifont], x, y, text, tlength);
    else
      XDrawString(XtDisplay(typewriter_drawing), XtWindow(typewriter_drawing), gc_table[ired][ifont], x, y, text, tlength);
  }
}

static void typewriter_update(int row1,int row2,int col1, int col2)
{
  int row,col;

  if(typewriter_wnd == NULL) return;

  if(row1<0) row1=0;
  if(row2>=LINEBUFFER) row2=LINEBUFFER-1;

  if(col1<0) col1=0;
  if(col2>=PAPERWIDTH) col2=PAPERWIDTH-1;

  for(row=row1; row<=row2; row++)
  for(col=col1; col<=col2; col++)
  typewriter_drawchar(row,col);
}

static void storechar(short c)
{
  int overwrite;
  if(redtext) c |= RED_MASK;
  overwrite=0;
  while(paper[cur_row][cur_col][overwrite]!=0 && overwrite<(MAXOVERWRITE-1) ) overwrite++;
  paper[cur_row][cur_col][overwrite] = c;
  cur_col++;
  if(cur_col >= PAPERWIDTH) cur_col=PAPERWIDTH-1;
}

static void typewriter_scroll()
{
  int i,j,k;

  for(i=0; i<LINEBUFFER-1; i++)
  for(j=0; j<PAPERWIDTH; j++)
  for(k=0; k<MAXOVERWRITE; k++)
  paper[i][j][k] = paper[i+1][j][k];

  for(j=0; j<PAPERWIDTH; j++)
  for(k=0; k<MAXOVERWRITE; k++)
  paper[LINEBUFFER-1][j][k] = 0;


  XCopyArea(XtDisplay(typewriter_drawing), XtWindow(typewriter_drawing), XtWindow(typewriter_drawing), scroll_gc,
  	    typewriter_CursorX(0), typewriter_CursorY(1),
	    typewriter_CursorX(PAPERWIDTH)-typewriter_CursorX(0), typewriter_CursorY(LINEBUFFER)-typewriter_CursorY(1),
	    typewriter_CursorX(0), typewriter_CursorY(0));

  in_copy++;
  typewriter_copywait();

  if(debug&DEBUGinterface) fprintf(debug_fh, "XCopyArea: (%d,%d) (%d,%d) (%d,%d)\n",
  	    typewriter_CursorX(0), typewriter_CursorY(1),
	    typewriter_CursorX(PAPERWIDTH)-typewriter_CursorX(0), typewriter_CursorY(LINEBUFFER)-typewriter_CursorY(1),
	    typewriter_CursorX(0), typewriter_CursorY(0));

            
}

void typewriter_SY(char c)
{
  short ch2;
  int this_col;
  int ccase;

  if(debug&DEBUGinterface) fprintf(debug_fh, "typewriter_SY: %d\n", c);
  if(c==64)
  {
    cur_col=0;
    cur_row++;
    if(cur_row==LINEBUFFER)
    {
      typewriter_scroll();
      cur_row--;
    }
  }
  else if(c<64)
  {
    ch2 = flx2a(c, &uppercase);
    this_col = cur_col;

    if(debug&DEBUGinterface) fprintf(debug_fh, "c: %d ch2: %d\n", (int) c, (int) ch2);
    
    if(c==29)
      redtext=1;
    else if(c==62)
      redtext=0;
    else
    {
      if(ch2==(unsigned char) '')
	ch2=218|SYMBOL_MASK;
      else if(ch2=='*')
	ch2=180|SYMBOL_MASK;
      else if(ch2=='&')
        ch2=217|SYMBOL_MASK;

      if(debug&DEBUGinterface) fprintf(debug_fh, "ch2: %d\n", (int) ch2);

      if(ch2 != 255) storechar(ch2);
      if(c==14) cur_col--;
    }
    if(typewriter_wnd == NULL) return;
    typewriter_drawchar(cur_row,this_col);
    XFlush(XtDisplay(typewriter_drawing));
  }
}

void typewriter_wait(int status)
{
  YE_wait=status;
  typewriter_waiting=status;
  YE_status();
}

static void typewriter_expose2_callback(Widget w, XtPointer client_data, XEvent *event, Boolean *cont)
{
  int row1, row2;
  int col1, col2;

  if(event->type == NoExpose)
  {
    row1 = 0;
    col1 = 0;
    row2 = LINEBUFFER-1;
    col2 = PAPERWIDTH-1;
  }
  else if(event->type == Expose || event->type == GraphicsExpose)
  {
    row1=event->xexpose.y/typewriter_font_height;
    row2=(event->xexpose.y+event->xexpose.height-1)/typewriter_font_height;
    col1=event->xexpose.x/typewriter_font_width;
    col2=(event->xexpose.x+event->xexpose.width-1)/typewriter_font_width;
  }
  typewriter_update(row1,row2,col1,col2);
  if(in_copy)
  {
    if((event->type == NoExpose) ||
       ((XExposeEvent *)event)->count == 0)
    {
      in_copy = 0;
    }
  }
}

static void typewriter_expose1_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  Boolean dummy;
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event = cbstruct->event;
  typewriter_expose2_callback(w, client_data, event, &dummy);
}

static void typewriter_copywait()
{
  XEvent reply;
  XEvent *rep = &reply;
  Boolean cont;
  /*
  	Wait for all events generated
	by XCopyArea.
  */

  while(in_copy)
  {
    cont = 1;
    XWindowEvent(XtDisplay(typewriter_drawing), XtWindow(typewriter_drawing), ExposureMask, &reply);
    switch(reply.type)
    {
    case Expose:
    case NoExpose:
    case GraphicsExpose:
      typewriter_expose2_callback(typewriter_drawing, NULL, &reply, &cont);
      break;
    }
  }
}

static void savechar(char c)
{
  if( ((c&128)==0) != (old_uppercase==0) )
  {
    if(typewriter_nchars<TYPEWRITER_TABLE_SIZE)
    {
      typewriter_char_table[typewriter_nchars] = ((c&128)==0)?58:60;
      typewriter_SY(typewriter_char_table[typewriter_nchars]);
      typewriter_nchars++;
    }
    else
      fprintf(stderr, "No room for case char.\n");
  }
  if(typewriter_nchars<TYPEWRITER_TABLE_SIZE)
  {
    typewriter_char_table[typewriter_nchars] = c&127;
    typewriter_SY(typewriter_char_table[typewriter_nchars]);
    typewriter_nchars++;
  }
  else
    fprintf(stderr, "No room char.\n");
}

static int uclc=0;

static void typewriter_input_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event = cbstruct -> event;
  int i,ilen,keycode;
  unsigned char ch2;

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "typewriter_input_callback called.\n");
    fprintf(debug_fh, "event type: %d\n", event->type);
  }

  if(event->type == KeyPress)
  {
    if(debug&DEBUGinterface) fprintf(debug_fh, "state: %d, keycode: %d\n", event->xkey.state, event->xkey.keycode);
    if(typewriter_waiting)
    {
      if(typewriter_nchars<(TYPEWRITER_TABLE_SIZE-1))
      {
        ilen=XLookupString((XKeyEvent *) event, strbuf, sizeof(strbuf), (KeySym *) &keycode, &compose);
	if(debug&DEBUGinterface) fprintf(debug_fh, "ilen: %d, keycode: %4.4x\n", ilen, keycode);
	if(keycode==XK_Shift_L || keycode==XK_Shift_R) uclc=1;
	for(i=0; i<ilen; i++)
	{
	  if(debug&DEBUGinterface) fprintf(debug_fh, "  character: %c\n", strbuf[i]);
	  uclc = 0;

	  old_uppercase = uppercase;
	  
	  ch2 = a2flx(strbuf[i], &uppercase);

	  if(ch2 == 255)
	  {
	    if(debug&DEBUGinterface) fprintf(debug_fh, "  Unknown character: %c\n", strbuf[i]);
	  }
	  else
	  {
	    savechar(ch2+128*uppercase);
	    YE_wait=0;
	    typewriter_wait(FALSE);
            usb_send_input("From typewriter_input_callback I");
	  }
	}
      }
      else
      {
        if(debug&DEBUGinterface) fprintf(debug_fh, "Character discarded; no room in typewriter_char_table\n");
      }
    }
    else
    {
      if(debug&DEBUGinterface) fprintf(debug_fh, "Character discarded; not waiting for input\n");
    }
  }
  else if(event->type == KeyRelease)
  {
    if(debug&DEBUGinterface) fprintf(debug_fh, "state: %d, keycode: %d\n", event->xkey.state, event->xkey.keycode);
    if(typewriter_waiting)
    {
      if(typewriter_nchars<(TYPEWRITER_TABLE_SIZE-2))
      {
        ilen=XLookupString((XKeyEvent *) event, strbuf, sizeof(strbuf), (KeySym *) &keycode, &compose);
	if(debug&DEBUGinterface) fprintf(debug_fh, "ilen: %d, keycode: %4.4x\n", ilen, keycode);
	if(keycode==XK_Shift_L || keycode==XK_Shift_R)
	{
	  if(uclc)
	  {
	    typewriter_char_table[typewriter_nchars++] = 60;
	    typewriter_char_table[typewriter_nchars++] = 58;
	    if(debug&DEBUGinterface) fprintf(debug_fh, "Generate UC/LC\n");
	    YE_wait=0;
	    uclc=0;
	    typewriter_wait(FALSE);
            usb_send_input("From typewriter_input_callback II");
	  }
	}
      }
    }
  }
}

static void typewriter_destroy_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  typewriter_wnd = NULL;

  update_viewmenu();
}

void typewriter_lowercase()
{
  /* Called by Reset/Mikrotempi stop */
  uppercase=0;
  typewriter_nchars=0;
  compose.compose_ptr = NULL;
  compose.chars_matched = FALSE;
}

void typewriter_init()
{
  int st;
  int i,j,k;
  double x;
  XGCValues values;
  XtGCMask valueMask;
  XmString stmp;
  Position dx,dy;
  int value_return, slider_size_return, increment_return, page_increment_return;
  int slider_max;

  if(typewriter_wnd != NULL)
  {
  }
  else
  {
    /* Get resources */

    XtGetApplicationResources(toplevel, &typewriter_applicationdata,
      typewriter_resources, XtNumber(typewriter_resources), NULL, 0);

    /*
    	Font administration:
    */

    typewriter_font_width = typewriter_applicationdata.textfont->max_bounds.width;
    typewriter_font_height = typewriter_applicationdata.textfont->ascent+typewriter_applicationdata.textfont->descent;
    typewriter_drawing_width=typewriter_font_width*PAPERWIDTH;
    typewriter_drawing_height=typewriter_font_height*LINEBUFFER;

    n=0;
    XtSetArg(wargs[n], XtNtitle, "GIER Typewriter"); n++;
    XtSetArg(wargs[n], XtNgeometry, "+10+400"); n++;
    typewriter_wnd = XtAppCreateShell("typewriter", "typewriter", topLevelShellWidgetClass, XtDisplay(toplevel), wargs, n);
    XtAddCallback(typewriter_wnd, XtNdestroyCallback, typewriter_destroy_callback, NULL);

    n=0;
    typewriter_form=XtCreateManagedWidget("typewriter_form", xmFormWidgetClass, typewriter_wnd, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNforeground, typewriter_applicationdata.lampcolor); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("     "))); n++;
    typewriter_input_label=XtCreateManagedWidget("typewriter_input_label", xmLabelWidgetClass, typewriter_form, wargs, n);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNtopWidget, typewriter_input_label); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNscrollBarPlacement, XmBOTTOM_LEFT); n++;
    XtSetArg(wargs[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
    XtSetArg(wargs[n], XmNwidth, typewriter_drawing_width+32); n++; /* Kludge */
    XtSetArg(wargs[n], XmNheight, 250); n++; /* Kludge */
    typewriter_swindow=XtCreateManagedWidget("typewriter_swindow", xmScrolledWindowWidgetClass, typewriter_form, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNhorizontalScrollBar, &typewriter_horScroll); n++;
    XtSetArg(wargs[n], XmNverticalScrollBar, &typewriter_verScroll); n++;
    XtGetValues(typewriter_swindow, wargs, n);

    XtUnmanageChild(typewriter_horScroll);

    n=0;
    XtSetArg(wargs[n], XmNwidth, typewriter_drawing_width); n++;
    XtSetArg(wargs[n], XmNheight, typewriter_drawing_height); n++;
    XtSetArg(wargs[n], XmNbackground, WhitePixel(XtDisplay(typewriter_swindow),XDefaultScreen(XtDisplay(typewriter_swindow)))); n++;
    typewriter_drawing=XtCreateManagedWidget("typewriter_drawing", xmDrawingAreaWidgetClass, typewriter_swindow, wargs, n);
    XtAddCallback(typewriter_drawing, XmNinputCallback, typewriter_input_callback, (XtPointer) NULL);
    XtAddCallback(typewriter_drawing, XmNexposeCallback, typewriter_expose1_callback, (XtPointer) NULL);
    XtAddEventHandler(typewriter_drawing, (EventMask) 0, TRUE, typewriter_expose2_callback, (XtPointer) NULL);

    /*
    	Create GC's:
    */

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.blackcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.textfont->fid;
    black_normal_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.redcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.textfont->fid;
    red_normal_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.blackcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.symbolfont->fid;
    black_symbol_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.redcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.symbolfont->fid;
    red_symbol_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.blackcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.subscriptfont->fid;
    black_subscript_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = typewriter_applicationdata.redcolor;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.font=typewriter_applicationdata.subscriptfont->fid;
    red_subscript_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    gc_table[0][0] = black_normal_gc;
    gc_table[1][0] = red_normal_gc;
    gc_table[0][1] = black_symbol_gc;
    gc_table[1][1] = red_symbol_gc;
    gc_table[0][2] = black_subscript_gc;
    gc_table[1][2] = red_subscript_gc;

    valueMask = GCBackground | GCGraphicsExposures;
    values.background = WhitePixel(XtDisplay(typewriter_drawing),XDefaultScreen(XtDisplay(typewriter_drawing)));
    values.graphics_exposures = (Bool) True;
    scroll_gc = XtGetGC(typewriter_drawing, valueMask, &values);

    XtRealizeWidget(typewriter_wnd);

    /* Scroll to bottom */

    n=0;
    XtSetArg(wargs[n], XmNmaximum, &slider_max); n++;
    XtGetValues(typewriter_verScroll, wargs, n);
    XmScrollBarGetValues(typewriter_verScroll, &value_return, &slider_size_return, &increment_return, &page_increment_return);
    value_return = slider_max-slider_size_return;
    XmScrollBarSetValues(typewriter_verScroll, value_return, slider_size_return, increment_return, page_increment_return, True);

    uppercase=0;
    cur_row=LINEBUFFER-1;
    cur_col=0;
    for(i=0; i<LINEBUFFER; i++)
    for(j=0; j<PAPERWIDTH; j++)
    for(k=0; k<MAXOVERWRITE; k++)
    paper[i][j][k] = 0;

    compose.compose_ptr = NULL;
    compose.chars_matched = FALSE;
  }
}

void typewriter_destroy()
{
  XtUnrealizeWidget(typewriter_wnd);
  XtDestroyWidget(typewriter_wnd);
}

int typewriter_is_visible()
{
  return(typewriter_wnd!=NULL);
}

typedef struct
{
  XFontStruct *textfont;
  Pixel tapebackground;
} reader_ApplicationData, *reader_ApplicationDataPtr;

static reader_ApplicationData reader_applicationdata;

static XtResource reader_resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(reader_ApplicationDataPtr, textfont),
   XtRString, "-adobe-courier-medium-r-*-*-*-180-75-75-*-*-*-*"
  },
  {"tapebackground", "Tapebackground", XtRPixel, sizeof(Pixel),
   XtOffset(reader_ApplicationDataPtr, tapebackground),
   XtRString, "gray"
  },
};

void reader_update()
{
  int i,j,k;
  char txt[READER_TAPE_WINDOW+READER_TAPE_OFFSET+1];
  unsigned char c;
  int mask;
  int reader_uppercase;

  if(reader_wnd == NULL) return;

  if(!reader_modified) return;

  txt[READER_TAPE_WINDOW+READER_TAPE_OFFSET] = '\0';

  if(debug&DEBUGinterface) fprintf(debug_fh, "reader_update: reader_nchars: %d\n", reader_nchars);

  for(j=0; j<READER_TAPE_WINDOW+READER_TAPE_OFFSET; j++)
  {
    if((j-READER_TAPE_OFFSET)>=reader_nchars)
    {
      c = ' ';
    }
    else
    {
      reader_uppercase = 0;
      c = flx2a(remove_parity((unsigned char) reader_char_table[j]), &reader_uppercase);
      if(c==255) c=' ';
    }
    txt[j] = c;
  }
  XDrawImageString(XtDisplay(reader_drawing), XtWindow(reader_drawing), reader_gray_gc, 0, reader_CursorY(0), txt, READER_TAPE_WINDOW+READER_TAPE_OFFSET);

  mask = 1;
  for(i=0; i<9; i++)
  {
    if(i==3)
    {
      for(j=0; j<READER_TAPE_WINDOW+READER_TAPE_OFFSET; j++) txt[j]='.';
    }
    else
    {
      for(j=0; j<READER_TAPE_WINDOW+READER_TAPE_OFFSET; j++)
      {
        if((j-READER_TAPE_OFFSET)>=reader_nchars)
	  k=0;
	else
	  k=reader_char_table[j];
	txt[j]= (k&mask)?'o':' ';
      }
      mask=mask<<1;
    }
    XDrawImageString(XtDisplay(reader_drawing), XtWindow(reader_drawing), reader_hole_gc, 0, reader_CursorY(9-i), txt, READER_TAPE_WINDOW+READER_TAPE_OFFSET);
    if(debug&DEBUGinterface) fprintf(debug_fh, "Print: %s\n", txt);
  }

  for(j=0; j<READER_TAPE_WINDOW+READER_TAPE_OFFSET; j++) txt[j]=' ';
  txt[READER_TAPE_OFFSET] = '^';
  XDrawImageString(XtDisplay(reader_drawing), XtWindow(reader_drawing), reader_gray_gc, 0, reader_CursorY(10), txt, READER_TAPE_WINDOW+READER_TAPE_OFFSET);

  reader_modified = False;
}

static void reader_update_buffer()
{
  int nread;
  unsigned char c1,c2;
  int i,old_uppercase;

  reader_modified = True;

  if(reader_nchars == READER_TAPE_WINDOW+READER_TAPE_OFFSET) return;
  if(reader_current_fd == -1) return;

  if(reader_filetype == 0)
  {
    nread=read(reader_current_fd, reader_char_table+reader_nchars, READER_TAPE_WINDOW+READER_TAPE_OFFSET-reader_nchars);

    if(nread <= 0)
    {
      close(reader_current_fd);
      reader_current_fd = -1;
      nread=0;
    }
    reader_nchars += nread;
  }
  else
  {
    /* read&convert ASCII tape */
    while(reader_nchars<(READER_TAPE_WINDOW+READER_TAPE_OFFSET))
    {
      if(reader_nsavedchars>0)
      {
	c2 = reader_savedchars[--reader_nsavedchars];
	nread=1;
      }
      else
      {
	nread=read(reader_current_fd, (char *) &c1, 1);
	if(nread <= 0)
	{
	  close(reader_current_fd);
	  reader_current_fd = -1;
	  nread=0;
	  goto ENDREAD;
	}
	else
	{
	  if(c1 == '\t')
	  {
	    do
	    {
	      reader_savedchars[reader_nsavedchars++] = 0;
	      reader_linepos++;
	    } while(reader_linepos%8!=0);
	    c2 = reader_savedchars[--reader_nsavedchars];
	    nread=1;
	  }
	  else
	  {
	    if(c1 == '#')
	    {
	      c2=0;
	      for(i=0; i<3; i++)
	      {
		nread=read(reader_current_fd, (char *) &c1, 1);
		if(nread <= 0)
		{
		  close(reader_current_fd);
		  reader_current_fd = -1;
		  nread=0;
		  goto ENDREAD;
		}
		if(c1>='0' && c1<='9') c2=c2*10+c1-'0';
	      }
	      c2 = c2&127;
	    }
	    else
	    {
	      old_uppercase = reader_uppercase;
	      c2 = a2flx(c1, &reader_uppercase);

	      if(old_uppercase != reader_uppercase)
	      {
		reader_savedchars[reader_nsavedchars++] = c2;
		c2 = reader_uppercase?60:58;
	      } /* if case shift */
	    } /* not #ddd */
	  } /* not TAB */
	} /* not end of file */
      } /* no chars saved */
      if(nread>0)
      {
	if(c2==64)
	{
	  reader_linepos=0;
	}
	else
	{
	  reader_linepos++;
	}
	reader_char_table[reader_nchars] = set_parity(c2);
      }
      reader_nchars += nread;
    } /* fill buffer */
  } /* ASCII read */

ENDREAD: ;

}

void reader_wait(int status)
{
  if(status)
  {
    YE_wait=1;
  }
  else
  {
    YE_wait=0;
    reader_update_buffer();
  }
  YE_status();
}

static void reader_fileopen_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;

  if(reader_current_fd != -1) close(reader_current_fd);

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);
  XtUnmanageChild(w);

  if(debug&DEBUGinterface) fprintf(debug_fh, "Opening file: %s\n", filename);

  reader_current_fd = open(filename, O_RDONLY);

  if(!strncmp(filename+strlen(filename)-3,"flx",3))
    reader_filetype=0;
  else
    reader_filetype=1;

  if(debug&DEBUGinterface) fprintf(debug_fh, "reader_filetype: %d\n", reader_filetype);

  reader_update_buffer();
  reader_modified = True;
  reader_update();
  if(reader_nchars>0)
  {
    if(usb_send_input("From reader_fileopen_callback"))
    {
      YE_wait=0;
      YE_status();
    }
  }

  XtFree(filename);
}

static void reader_fileclose_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnmanageChild(w);
}

static void pread_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionDoSearch(reader_filedialog, NULL);
  XtManageChild(reader_filedialog);
}

static void reset_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i;
  reader_nchars=READER_TAPE_OFFSET;

  for(i=0; i<READER_TAPE_OFFSET; i++) reader_char_table[i]=0;

  pread_callback(w, client_data, call_data);
}

static void skip_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
}

static void up_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
}

static void reader_expose_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event=cbstruct->event;
  int row1,row2;

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "reader_expose_callback called. Event type: %d\n", event->type);
    fprintf(debug_fh, "x: %d, y: %d\n", event->xexpose.x, event->xexpose.y);
    fprintf(debug_fh, "width: %d, height: %d\n", event->xexpose.width, event->xexpose.height);
  }

  reader_modified = True;
  reader_update();
}

static void reader_input_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event = cbstruct -> event;
  int i,ilen,keycode;

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "reader_input_callback called.\n");
    fprintf(debug_fh, "event type: %d\n", event->type);
  }

}

static void reader_destroy_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  reader_wnd = NULL;

  update_viewmenu();
}

void reader_init()
{
  int st;
  int i,j,k;
  double x;
  XGCValues values;
  XtGCMask valueMask;
  XmString stmp,stmp2;
  Position dx,dy;

  if(reader_wnd != NULL)
  {
  }
  else
  {
    /* Get resources */

    XtGetApplicationResources(toplevel, &reader_applicationdata,
      reader_resources, XtNumber(reader_resources), NULL, 0);

    /*
        Font administration:
    */

    reader_font_width = reader_applicationdata.textfont->max_bounds.width;
    reader_font_height = reader_applicationdata.textfont->ascent+reader_applicationdata.textfont->descent;
    reader_drawing_width=reader_font_width*(READER_TAPE_WINDOW+READER_TAPE_OFFSET);
    reader_drawing_height=reader_font_height*11;

    n=0;
    XtSetArg(wargs[n], XtNtitle, "GIER Paper Tape Reader"); n++;
    XtSetArg(wargs[n], XtNgeometry, "+10+110"); n++;
    reader_wnd = XtAppCreateShell("reader", "reader", topLevelShellWidgetClass, XtDisplay(toplevel), wargs, n);
    XtAddCallback(reader_wnd, XtNdestroyCallback, reader_destroy_callback, NULL);

    n=0;
    reader_form=XtCreateManagedWidget("reader_form", xmFormWidgetClass, reader_wnd, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("RESET"))); n++;
    reader_reset = XtCreateManagedWidget("reader_reset", xmPushButtonWidgetClass, reader_form, wargs, n);
    XtAddCallback(reader_reset, XmNactivateCallback, reset_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNleftWidget, reader_reset); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("READ"))); n++;
    reader_pread = XtCreateManagedWidget("reader_pread", xmPushButtonWidgetClass, reader_form, wargs, n);
    XtAddCallback(reader_pread, XmNactivateCallback, pread_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNleftWidget, reader_pread); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("SKIP"))); n++;
    reader_skip = XtCreateManagedWidget("reader_skip", xmPushButtonWidgetClass, reader_form, wargs, n);
    XtAddCallback(reader_skip, XmNactivateCallback, skip_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNleftWidget, reader_skip); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("UP"))); n++;
    reader_up = XtCreateManagedWidget("reader_up", xmPushButtonWidgetClass, reader_form, wargs, n);
    XtAddCallback(reader_up, XmNactivateCallback, up_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNtopWidget, reader_reset); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNwidth, reader_drawing_width); n++;
    XtSetArg(wargs[n], XmNheight, reader_drawing_height); n++;
    XtSetArg(wargs[n], XmNbackground, reader_applicationdata.tapebackground); n++;
    reader_drawing=XtCreateManagedWidget("reader_drawing", xmDrawingAreaWidgetClass, reader_form, wargs, n);
    XtAddCallback(reader_drawing, XmNexposeCallback, reader_expose_callback, (XtPointer) NULL);
    XtAddCallback(reader_drawing, XmNinputCallback, reader_input_callback, (XtPointer) NULL);

    n=0;
    XtSetArg(wargs[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
    XtSetArg(wargs[n], XmNautoUnmanage, False); n++;
    XtSetArg(wargs[n], XmNnoResize, TRUE); n++;
    XtSetArg(wargs[n], XmNresizePolicy, XmRESIZE_NONE); n++;
    XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*"))); n++;
    XtSetArg(wargs[n], XmNdialogTitle, (stmp2=XmStringCreateSimple("Reader File Selection"))); n++;

    reader_filedialog = XmCreateFileSelectionDialog(reader_form, "FileDialog", wargs, n);
    XmStringFree(stmp);
    XmStringFree(stmp2);

    XtAddCallback(reader_filedialog, XmNokCallback, reader_fileopen_callback, (XtPointer)NULL);
    XtAddCallback(reader_filedialog, XmNcancelCallback, reader_fileclose_callback, (XtPointer)NULL);

    /*
    	Create GC's:

	The paper is white, holes are black, and the background is gray:
    */

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = BlackPixel(XtDisplay(reader_drawing),XDefaultScreen(XtDisplay(reader_drawing)));
    values.background = WhitePixel(XtDisplay(reader_drawing),XDefaultScreen(XtDisplay(reader_drawing)));
    values.font=reader_applicationdata.textfont->fid;
    reader_hole_gc = XtGetGC(reader_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = BlackPixel(XtDisplay(reader_drawing),XDefaultScreen(XtDisplay(reader_drawing)));
    values.background = reader_applicationdata.tapebackground;
    values.font=reader_applicationdata.textfont->fid;
    reader_gray_gc = XtGetGC(reader_drawing, valueMask, &values);

    XtRealizeWidget(reader_wnd);

  }
}

void reader_destroy()
{
  XtUnrealizeWidget(reader_wnd);
  XtDestroyWidget(reader_wnd);
}

int reader_is_visible()
{
  return(reader_wnd!=NULL);
}


static void printer_setfilename(char *filename)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(filename))); n++;
  XtSetValues(printer_psfilename, wargs, n);

  XmStringFree(stmp);
}
  
static void printer_update()
{
}

static void printer_papersize_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *s;

  s = XmTextFieldGetString(lines_edit);
  sscanf(s, "%d", &lines);
  XFree(s);

  letter = (long) client_data;
  if(letter && (lines==67))
  {
    lines = 63;
  }
  else if((!letter) && (lines==63))
  {
    lines = 67;
  }
  sprintf(lines_buffer, "%d", lines);
  XmTextFieldSetString(lines_edit, lines_buffer);
}

static void printer_filedialog_create()
{
  XmString stmp;
  Widget papersize_label, papersize_pane,
         lpi_edit_label, fontsize_edit_label, lines_edit_label, uadvance_select_label;
  Widget A4_w, Letter_w;

  n=0;
  XtSetArg(wargs[n], XtNtitle, "PostScript File Selection"); n++;
  printer_filedialog = XtAppCreateShell("printer_filedialog", "printer_filedialog", topLevelShellWidgetClass, XtDisplay(toplevel), wargs, n);

  n=0;
  printer_fileform = XtCreateManagedWidget("form", xmFormWidgetClass, printer_filedialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*.ps"))); n++;
  printer_fileselection = XtCreateManagedWidget("printer_fileselection", xmFileSelectionBoxWidgetClass, printer_fileform, wargs, n);
  XmStringFree(stmp);

  XtAddCallback(printer_fileselection, XmNokCallback, printer_fileopen_callback, (XtPointer)NULL);
  XtAddCallback(printer_fileselection, XmNcancelCallback, printer_fileclose_callback, (XtPointer)NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  papersize_label = XtCreateManagedWidget("Select paper size", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 0); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  papersize = XmCreateOptionMenu(printer_fileform, "papersize", wargs, n);

  n=0;
  papersize_pane = XmCreatePulldownMenu(papersize, "pulldown", wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("A4"))); n++;
  A4_w = XtCreateManagedWidget("A4_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(A4_w, XmNactivateCallback, printer_papersize_callback, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("Letter"))); n++;
  Letter_w = XtCreateManagedWidget("letter_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(Letter_w, XmNactivateCallback, printer_papersize_callback, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNsubMenuId, papersize_pane); n++;
  XtSetValues(papersize, wargs, n);

  XtManageChild(papersize);

  n=0;
  XtSetArg(wargs[n], XmNmenuHistory, letter?Letter_w:A4_w); n++;
  XtSetValues(papersize, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  lpi_edit_label = XtCreateManagedWidget("Select lines per inch", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, lpi_edit_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNvalue, lpi_buffer); n++;
  sprintf(lpi_buffer, "%d", lpi);
  XtSetArg(wargs[n], XmNmaxLength, sizeof(lpi_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(lpi_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  lpi_edit = XtCreateManagedWidget("lpi_edit", xmTextFieldWidgetClass, printer_fileform, wargs, n);
  
  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, lpi_edit); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  fontsize_edit_label = XtCreateManagedWidget("Select font size", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, fontsize_edit_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNvalue, fontsize_buffer); n++;
  sprintf(fontsize_buffer, "%lg", fontsize);
  XtSetArg(wargs[n], XmNmaxLength, sizeof(fontsize_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(fontsize_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  fontsize_edit = XtCreateManagedWidget("fontsize_edit", xmTextFieldWidgetClass, printer_fileform, wargs, n);
  
  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, fontsize_edit); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  lines_edit_label = XtCreateManagedWidget("Select lines per page", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, lines_edit_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNvalue, lines_buffer); n++;
  sprintf(lines_buffer, "%d", lines);
  XtSetArg(wargs[n], XmNmaxLength, sizeof(lines_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(lines_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  lines_edit = XtCreateManagedWidget("lines_edit", xmTextFieldWidgetClass, printer_fileform, wargs, n);
  
  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, lines_edit); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  uadvance_select_label = XtCreateManagedWidget("Print _ and | correctly", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, uadvance_select_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  uadvance_check = XmCreateSimpleCheckBox(printer_fileform, "uadvance_check", wargs, n);
  XtManageChild(uadvance_check);

  n=0;
#if XmVERSION==1
  XtSetArg(wargs[n], XmNindicatorOn, True); n++;
  XtSetArg(wargs[n], XmNset, uadvance?True:False); n++;
#else
  XtSetArg(wargs[n], XmNindicatorOn, XmINDICATOR_CHECK_BOX); n++;
  XtSetArg(wargs[n], XmNset, uadvance?XmSET:XmUNSET); n++;
#endif
  XtSetArg(wargs[n], XmNindicatorSize, 20); n++;
  uadvance_select = XmCreateToggleButton(uadvance_check, "", wargs, n);
  XtManageChild(uadvance_select);


  XtRealizeWidget(printer_filedialog);

  XmFileSelectionDoSearch(printer_fileselection, NULL);
}

static void printer_filedialog_close()
{
  XtDestroyWidget(printer_filedialog);
}

void printer_SY(char c)
{
  while(!is_printing)
  {
    if(printer_wnd == NULL) printer_init();
    printer_filedialog_create();
    while((!is_printing) || XtAppPending(app))
    {
      XtAppProcessEvent(app,XtIMAll);
    }
  }

  ps_SY((unsigned char) c);

  printer_update();
}

static void printer_fileopen_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;
  char *s;

  if(is_printing)
  {
    ps_close();
    is_printing=0;
  }

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);

  /* Get parameters */

  s = XmTextFieldGetString(lpi_edit);
  sscanf(s, "%d", &lpi);
  XFree(s);

  s = XmTextFieldGetString(fontsize_edit);
  sscanf(s, "%lg", &fontsize);
  XFree(s);

  s = XmTextFieldGetString(lines_edit);
  sscanf(s, "%d", &lines);
  XFree(s);

  n=0;
  uadvance=0;
  XtSetArg(wargs[n], XmNset, &uadvance); n++;
  XtGetValues(uadvance_select, wargs, n);

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "lpi: %d\n", lpi);
    fprintf(debug_fh, "fontsize: %lg\n", fontsize);
    fprintf(debug_fh, "lines: %d\n", lines);
    fprintf(debug_fh, "uadvance: %d\n", uadvance);
    fprintf(debug_fh, "Opening file: %s\n", filename);
  }

  printer_filedialog_close();

  is_printing = ps_open(filename, letter, fontsize, lpi, lines, !uadvance);

  if(is_printing) printer_setfilename(filename);
  XtFree(filename);
}

static void printer_fileclose_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  printer_filedialog_close();
}

static void printer_endfile_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i;

  if(is_printing)
  {
    ps_close();
    is_printing = 0;
    printer_setfilename("- no file -");
  }
  printer_update();
}

static void printer_newfile_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  printer_endfile_callback(w, client_data, call_data);
  printer_filedialog_create();
}

static void printer_destroy_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  printer_wnd = NULL;

  update_viewmenu();
}


void printer_init()
{
  int st;
  int i,j,k;
  double x;
  XmString stmp,stmp2;

  if(printer_wnd != NULL)
  {
  }
  else
  {
    n=0;
    XtSetArg(wargs[n], XtNtitle, "GIER Lineprinter"); n++;
    XtSetArg(wargs[n], XtNgeometry, "+230+10"); n++;
    printer_wnd = XtAppCreateShell("printer", "printer", topLevelShellWidgetClass, XtDisplay(toplevel), wargs, n);
    XtAddCallback(printer_wnd, XtNdestroyCallback, printer_destroy_callback, NULL);

    n=0;
    printer_form=XtCreateManagedWidget("printer_form", xmFormWidgetClass, printer_wnd, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("New File"))); n++;
    printer_newfile = XtCreateManagedWidget("printer_newfile", xmPushButtonWidgetClass, printer_form, wargs, n);
    XtAddCallback(printer_newfile, XmNactivateCallback, printer_newfile_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNleftWidget, printer_newfile); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("End File"))); n++;
    printer_endfile = XtCreateManagedWidget("printer_endfile", xmPushButtonWidgetClass, printer_form, wargs, n);
    XtAddCallback(printer_endfile, XmNactivateCallback, printer_endfile_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNtopWidget, printer_newfile); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("- no file -"))); n++;
    printer_psfilename = XtCreateManagedWidget("printer_psfilename", xmLabelWidgetClass, printer_form, wargs, n);
    XmStringFree(stmp);

    XtRealizeWidget(printer_wnd);

  }
}

void printer_destroy()
{
  XtUnrealizeWidget(printer_wnd);
  XtDestroyWidget(printer_wnd);
}

int printer_is_visible()
{
  return(printer_wnd!=NULL);
}

/* Local variables */

typedef struct
{
  XFontStruct *textfont;
  Pixel tapebackground;
} punch_ApplicationData, *punch_ApplicationDataPtr;

static punch_ApplicationData punch_applicationdata;

static XtResource punch_resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(punch_ApplicationDataPtr, textfont),
   XtRString, "-adobe-courier-medium-r-*-*-*-180-75-75-*-*-*-*"
  },
  {"tapebackground", "Tapebackground", XtRPixel, sizeof(Pixel),
   XtOffset(punch_ApplicationDataPtr, tapebackground),
   XtRString, "gray"},
};

void punch_init();

static void punch_settapename(char *filename)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(filename))); n++;
  XtSetValues(punch_tapename, wargs, n);

  XmStringFree(stmp);
}
  
static void punch_update()
{
  int i,j,k;
  char txt[PUNCH_TAPE_WINDOW+1];
  unsigned char c;
  int mask;
  int punch_uppercase;

  if(punch_wnd == NULL) return;

  txt[PUNCH_TAPE_WINDOW] = '\0';

  for(j=0; j<PUNCH_TAPE_WINDOW; j++)
  {
    punch_uppercase = 0;
    c = flx2a(remove_parity((unsigned char) punch_char_table[j]), &punch_uppercase);
    if(c==255) c=' ';
    txt[j] = c;
  }
  XDrawImageString(XtDisplay(punch_drawing), XtWindow(punch_drawing), punch_gray_gc, 0, punch_CursorY(0), txt, PUNCH_TAPE_WINDOW);

  mask = 1;
  for(i=0; i<9; i++)
  {
    if(i==3)
    {
      for(j=0; j<PUNCH_TAPE_WINDOW; j++) txt[j]='.';
    }
    else
    {
      for(j=0; j<PUNCH_TAPE_WINDOW; j++)
      {
	k=punch_char_table[j];
	txt[j]= (k&mask)?'o':' ';
      }
      mask=mask<<1;
    }
    XDrawImageString(XtDisplay(punch_drawing), XtWindow(punch_drawing), punch_hole_gc, 0, punch_CursorY(9-i), txt, PUNCH_TAPE_WINDOW);
    if(debug&DEBUGinterface) fprintf(debug_fh, "Print: %s\n", txt);
  }

  for(j=0; j<PUNCH_TAPE_WINDOW; j++) txt[j]=' ';
  txt[PUNCH_TAPE_WINDOW-1] = '^';
  XDrawImageString(XtDisplay(punch_drawing), XtWindow(punch_drawing), punch_gray_gc, 0, punch_CursorY(10), txt, PUNCH_TAPE_WINDOW);
}

void punch_SY(char c)
{
  int st,i;
  unsigned char c2;
  char noparc;
  char hashbuf[5];

  noparc = c;
  c = set_parity(c);

  while(punch_current_fd == -1) 
  {
    if(punch_wnd == NULL) punch_init();
    XmFileSelectionDoSearch(punch_filedialog, NULL);
    XtManageChild(punch_filedialog);
    while(punch_current_fd == -1 || XtAppPending(app))
    {
      XtAppProcessEvent(app,XtIMAll);
    }
  }

  if(punch_filetype == 0)
  {
    st=write(punch_current_fd, &c, 1);
  }
  else
  {
    c2 = flx2a(noparc, &punch_uppercase);
    if(c2 == 255)
    {
      if(noparc != 58 && noparc != 60)
      {
        sprintf(hashbuf, "#%3.3d", (int) noparc);
	st=write(punch_current_fd, hashbuf, 4);
      }
    }
    else
    {
      st=write(punch_current_fd, &c2, 1);
    }
  }

  if(st == -1)
  {
    close(punch_current_fd);
    punch_current_fd = -1;
    punch_settapename("- no tape -");
  }

  for(i=0; i<(PUNCH_TAPE_WINDOW-1); i++)
    punch_char_table[i] = punch_char_table[i+1];

  punch_char_table[PUNCH_TAPE_WINDOW-1] = c;

  punch_update();
}

static void punch_fileopen_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;

  if(punch_current_fd != -1) close(punch_current_fd);

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);
  XtUnmanageChild(w);

  if(debug&DEBUGinterface) fprintf(debug_fh, "Opening file: %s\n", filename);

  punch_current_fd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0666);
  if(!strncmp(filename+strlen(filename)-3,"flx",3))
    punch_filetype = 0;
  else
    punch_filetype = 1;

  if(punch_current_fd != -1) punch_settapename(filename);
  XtFree(filename);
}

static void punch_fileclose_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnmanageChild(w);
}

static void endtape_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i;

  for(i=0; i<PUNCH_TAPE_WINDOW; i++) punch_char_table[i]=0;
  if(punch_current_fd != -1)
  {
    close(punch_current_fd);
    punch_current_fd = -1;
    punch_settapename("- no tape -");
  }
  punch_update();
}

static void newtape_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  endtape_callback(w, client_data, call_data);
  XmFileSelectionDoSearch(punch_filedialog, NULL);
  XtManageChild(punch_filedialog);
}

static void punch_expose_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event=cbstruct->event;

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "punch_expose_callback called. Event type: %d\n", event->type);
    fprintf(debug_fh, "x: %d, y: %d\n", event->xexpose.x, event->xexpose.y);
    fprintf(debug_fh, "width: %d, height: %d\n", event->xexpose.width, event->xexpose.height);
  }

  punch_update();
}

static void punch_input_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cbstruct = (XmDrawingAreaCallbackStruct *) call_data;
  XEvent *event = cbstruct -> event;
  int i,ilen,keycode;

  if(debug&DEBUGinterface)
  {
    fprintf(debug_fh, "punch_input_callback called.\n");
    fprintf(debug_fh, "event type: %d\n", event->type);
  }

}

static void punch_destroy_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
  punch_wnd = NULL;

  update_viewmenu();
}

void punch_init()
{
  int st;
  int i,j,k;
  double x;
  XGCValues values;
  XtGCMask valueMask;
  Colormap cmap;
  XColor ignore;
  XmString stmp,stmp2;
  Position dx,dy;

  if(punch_wnd != NULL)
  {
  }
  else
  {
    /* Get resources */

    XtGetApplicationResources(toplevel, &punch_applicationdata,
      punch_resources, XtNumber(punch_resources), NULL, 0);

    /*
        Font administration:
    */

    punch_font_width = punch_applicationdata.textfont->max_bounds.width;
    punch_font_height = punch_applicationdata.textfont->ascent+punch_applicationdata.textfont->descent;
    punch_drawing_width=punch_font_width*PUNCH_TAPE_WINDOW;
    punch_drawing_height=punch_font_height*11;

    n=0;
    XtSetArg(wargs[n], XtNtitle, "GIER Paper Tape Punch"); n++;
    XtSetArg(wargs[n], XtNgeometry, "+385+110"); n++;
    punch_wnd = XtAppCreateShell("punch", "punch", topLevelShellWidgetClass, XtDisplay(toplevel), wargs, n);
    XtAddCallback(punch_wnd, XtNdestroyCallback, punch_destroy_callback, NULL);

    n=0;
    punch_form=XtCreateManagedWidget("punch_form", xmFormWidgetClass, punch_wnd, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("New Tape"))); n++;
    punch_newtape = XtCreateManagedWidget("punch_newtape", xmPushButtonWidgetClass, punch_form, wargs, n);
    XtAddCallback(punch_newtape, XmNactivateCallback, newtape_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNleftWidget, punch_newtape); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("End Tape"))); n++;
    punch_endtape = XtCreateManagedWidget("punch_endtape", xmPushButtonWidgetClass, punch_form, wargs, n);
    XtAddCallback(punch_endtape, XmNactivateCallback, endtape_callback, NULL);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNtopWidget, punch_newtape); n++;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple("- no tape -"))); n++;
    punch_tapename = XtCreateManagedWidget("punch_tapename", xmLabelWidgetClass, punch_form, wargs, n);
    XmStringFree(stmp);

    n=0;
    XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(wargs[n], XmNtopWidget, punch_tapename); n++;
    XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(wargs[n], XmNwidth, punch_drawing_width); n++;
    XtSetArg(wargs[n], XmNheight, punch_drawing_height); n++;
    XtSetArg(wargs[n], XmNbackground, punch_applicationdata.tapebackground); n++;
    punch_drawing=XtCreateManagedWidget("punch_drawing", xmDrawingAreaWidgetClass, punch_form, wargs, n);
    XtAddCallback(punch_drawing, XmNexposeCallback, punch_expose_callback, (XtPointer) NULL);
    XtAddCallback(punch_drawing, XmNinputCallback, punch_input_callback, (XtPointer) NULL);

    n=0;
    XtSetArg(wargs[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
    XtSetArg(wargs[n], XmNautoUnmanage, False); n++;
    XtSetArg(wargs[n], XmNnoResize, TRUE); n++;
    XtSetArg(wargs[n], XmNresizePolicy, XmRESIZE_NONE); n++;
    XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*"))); n++;
    XtSetArg(wargs[n], XmNdialogTitle, (stmp2=XmStringCreateSimple("Punch File Selection"))); n++;

    punch_filedialog = XmCreateFileSelectionDialog(punch_form, "FileDialog", wargs, n);
    XmStringFree(stmp);
    XmStringFree(stmp2);

    XtAddCallback(punch_filedialog, XmNokCallback, punch_fileopen_callback, (XtPointer)NULL);
    XtAddCallback(punch_filedialog, XmNcancelCallback, punch_fileclose_callback, (XtPointer)NULL);

    /*
    	Create GC's:

	The paper is white, holes are black, and the background is gray:
    */

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = BlackPixel(XtDisplay(punch_drawing),XDefaultScreen(XtDisplay(punch_drawing)));
    values.background = WhitePixel(XtDisplay(punch_drawing),XDefaultScreen(XtDisplay(punch_drawing)));
    values.font=punch_applicationdata.textfont->fid;
    punch_hole_gc = XtGetGC(punch_drawing, valueMask, &values);

    valueMask = GCForeground | GCBackground | GCFont;
    values.foreground = BlackPixel(XtDisplay(punch_drawing),XDefaultScreen(XtDisplay(punch_drawing)));
    values.background = punch_applicationdata.tapebackground;
    values.font=punch_applicationdata.textfont->fid;
    punch_gray_gc = XtGetGC(punch_drawing, valueMask, &values);

    for(i=0; i<PUNCH_TAPE_WINDOW; i++) punch_char_table[i]=0;
    punch_current_fd = -1;

    XtRealizeWidget(punch_wnd);

  }
}

void punch_destroy()
{
  XtUnrealizeWidget(punch_wnd);
  XtDestroyWidget(punch_wnd);
}

int punch_is_visible()
{
  return(punch_wnd != NULL);
}

main(int argc, char **argv)
{
  interface_init(argc, argv);
  GIERmain();
}
